/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* ************************************************************
 *   problem: *   u_{t} - \Delta{u} = func_f  (x, t) \in \Omega X (0, T)
 *   u = func_g (x, t) \in \partial\Omega X [0, T]
 *   u = func_u0       x \in \Omega  t = 0.0
 *   
 * $Id: heat.c,v 1.32 2007/09/27 11:50:33 zlb Exp $
 **************************************************************/
#include "phg.h"
#include <math.h>
#include <stdlib.h>

#define OUTPUT_MESH 0   /* int .0: won't output; > 0 output */

/* these parameters can be modified */
#define tol_space  0.01
#define tol_time   tol_space
#define T          1            /* time domain: [0, T],T > 0 */
#define FLAGA   25              /* parameter used by func_f, func_g, > 0. */
#define FLAGB   0.7             /* parameter used by func_f, func_g, > 0. */

/* parameters used for controling refining and coarsening */
/* Can be modified */
/* parameters for max strategy */
#define refine_max     0.8
#define osc_max        0.8
#define coarsen_max    0.05

/* default time step */
/* can be modified */
static FLOAT stime = 0.1;

/* thest can't be modified */
static FLOAT ctime = 0.0;       /* current time. Modified by program */
static FLOAT ptime = 0.0;       /* previous time. */
static FLOAT SpaceTOL = tol_space / T;         /* TOL / T, */
static FLOAT TimeTOL = tol_time / (2.0 * T) ;  /* TOL / (2 * T) */
static FLOAT FTOL;                             /* Sqrt(TOL) / (2 * T) */

static void     /* rhs: f */
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT tmp, etmp, ct, st;
    ct = Cos(8.0 * M_PI * ctime);
    st = Sin(8.0 * M_PI * ctime);
    tmp = (x - 0.5 - 0.4 * st) * (x - 0.5 - 0.4 * st)
        + (y - 0.5 - 0.4 * ct) * (y - 0.5 - 0.4 * ct)
        + (z - 1.0) * (z - 1.0);
    etmp = 1.0 / (FLAGA * tmp + FLAGB + 0.2);
    *value = ((6.4 * FLAGA * M_PI * ct * (x - 0.5 - 0.4 * st) 
             - 6.4 * FLAGA * M_PI * st * (y - 0.5 - 0.4 * ct)) * etmp * etmp
             - 4.0 * FLAGA * FLAGA * tmp * etmp * etmp * etmp * etmp
	     - 8.0 * FLAGA * FLAGA * tmp * etmp * etmp * etmp
	     + 6.0 * FLAGA * etmp * etmp) * Exp(etmp) * Exp(-2.5);
}

static void    /* boundary function: Dirichlet */
func_g(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT tmp;
    tmp = 1.0 / (FLAGA * ((x - 0.5 - 0.4 * Sin(8.0 * M_PI * ctime)) 
			    * (x - 0.5 - 0.4 * Sin(8.0 * M_PI * ctime))
                        + (y - 0.5 - 0.4 * Cos(8.0 * M_PI * ctime)) 
             	            * (y - 0.5 - 0.4 * Cos(8.0 * M_PI * ctime))
                        + (z - 1.0) * (z - 1.0)) + FLAGB + 0.2);
    *value = exp(tmp) * Exp(-2.5);
}

static void /* initial solution u0: t = 0*/
func_u0(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT tmp;
    tmp = 1.0 / (FLAGA * ((x - 0.5) * (x - 0.5)
			+ (y - 0.9) * (y - 0.9)
                        + (z - 1.0) * (z - 1.0)) + FLAGB + 0.2);
    *value = exp(tmp) * Exp(-2.5);
}

static double
cal_time(BOOLEAN flag)
{
    static double t0 = 0.0;
    double et, tt[3];

    phgGetTime(tt);
    et = tt[2] - t0;
    t0 = tt[2];

    if (flag){
        phgPrintf("  %0.4lfs",et);
    }

    return et;
}

/* mark elements for refining and coarsening */
static void 
MarkElements(DOF *error, DOF *osc)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT flag, flag2, oscmax, PE_error;
    FLOAT tmpe, tmpo;
    long  countr, countc;
   
    /* clean marks */
    ForAllElements(g, e){
        e->mark = 0;
    }
    
    PE_error = phgDofNormInftyVec(error);
    oscmax = phgDofNormInftyVec(osc);
    /* mark for refining */
    /* space indicator */
    flag = PE_error * refine_max * refine_max;
    flag2 = oscmax * osc_max * osc_max;
    countr = countc = 0;
    ForAllElements(g, e){
        tmpe = *DofElementData(error, e->index);
	tmpo = *DofElementData(osc, e->index);
        if (tmpe > flag || tmpo > flag2){
		e->mark = 1;
		countr++;
	}
    }
		
    countc = countr;
    phgPrintf("Elements marked for refining:%ld\n", countc);
    
    /* mark for coarsening */
    /* space indicator only */
    flag = PE_error * coarsen_max * coarsen_max;
    countr = countc = 0;
    ForAllElements(g, e){
        if (*DofElementData(error, e->index) < flag){
		e->mark = -1;
		countc++;
	}
    }

    countr = countc;
    phgPrintf("Elements marked for coarsening:%ld\n", countr);

}

/* update f_h */
static void
update_fh(DOF *f_h, DOF *ftmp)
{
    phgDofSetDataByValue(f_h, 0.0);
    phgDofSetDataByFunction(ftmp, func_f);
    phgDofAXPY(1.0 / 6.0, ftmp, &f_h);
    ctime -= stime * 0.5;
    phgDofSetDataByFunction(ftmp, func_f);
    phgDofAXPY(2.0 / 3.0, ftmp, &f_h);
    ctime -= stime * 0.5;
    phgDofSetDataByFunction(ftmp, func_f);
    phgDofAXPY(1.0 / 6.0, ftmp, &f_h);
    ctime += stime;
}

/* build linear systme */
static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f, DOF *u_p)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    int dim = u_h->dim;         /* DOF dimension */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j, k, n;
    FLOAT a[N][N], A[N][dim], B[N], buffer[N], *row;
    INT I[N][dim], J[N][dim];
    FLOAT tmp;
   
    ForAllElements(g, e) {
	for (i = 0; i < N; i++) {
	    for (k = 0; k < dim; k++)
		I[i][k] = phgSolverMapE2L(solver, 0, e, i * dim + k);
	    for (j = 0; j < i; j++) {
		a[j][i] = a[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT) * stime
		    + phgQuadBasDotBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	    }
	    a[i][i] = phgQuadGradBasDotGradBas(e, u_h, i, u_h, i, QUAD_DEFAULT) * stime
		    + phgQuadBasDotBas(e, u_h, i, u_h, i, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, func_g, buffer, &B[i],
				DOF_PROJ_NONE)) {
		row = buffer;
	    }
	    else if (phgDofGetElementBoundaryType(u_h, e, i * dim) & NEUMANN) {
		/* TODO */
		row = NULL;
		phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    }
	    else {
		/* right hand side: \int f * phi_i */
		phgQuadDofTimesBas(e, f, u_h, i, QUAD_DEFAULT, &B[i]);
		B[i] = stime * B[i];
		phgQuadDofTimesBas(e, u_p, u_h, i, QUAD_DEFAULT, &tmp);
		B[i] += tmp;
		row = a[i];
	    }
	    for (j = 0, n = 0; j < N; j++) {
		for (k = 0; k < dim; k++) {
		    A[n][k] = row[j];
		    J[n++][k] = I[j][k];
		}
	    }
	    phgSolverAddMatrixEntries(solver, 1, I[i], n, J[0], A[0]); 
	}
	phgSolverAddRHSEntries(solver, N * dim, I[0], &B[0]);
    }
}

static void  /* solve Au = b */
SolveLinearSystem(OEM_SOLVER *sol, DOF *u_h, DOF *f_h, DOF *u_p)
{
    GRID *g = u_h->g;
    SOLVER *solver;
    int nits;
    phgPrintf("Number of DOF:           %d\n", DofGetDataCountGlobal(u_h));
    phgPrintf("Number of elements:      %d\n", g->nleaf_global);
    phgPrintf("Load imbalance:          %lg\n", (double)g->lif);
                       
    phgPrintf("create solver:         ");
    cal_time(FALSE);
    solver = phgSolverCreate(sol, u_h, NULL);
    cal_time(TRUE);
    phgPrintf("\n");
                
    phgPrintf("build linear system:   ");
    cal_time(FALSE);
    build_linear_system(solver, u_h, f_h, u_p);
    cal_time(TRUE);
    phgPrintf("\n");
              
    phgPrintf("solve solver:          ");
    cal_time(FALSE);
    nits = phgSolverSolve(solver, TRUE, u_h, NULL);
    cal_time(TRUE);
    phgPrintf("     [%d]    \n", nits);
    phgSolverDestroy(&solver);

}

static FLOAT    /* time residual, global */
estimate_time_error(DOF *u_h, DOF *u_p)
{
    DOF *tmp, *grad;
    FLOAT ss;

    tmp = phgDofCopy(u_h, NULL, DOF_DEFAULT, "tmp2");
    phgDofAXPY( -1.0, u_p, &tmp);
    grad = phgDofGradient(tmp, NULL, NULL, "grad");
    ss = phgDofNormL2(grad);

    phgDofFree(&tmp);
    phgDofFree(&grad);

    return ss * ss * (FLOAT)(1.L/3.L);
}

/* global source error, use Simpson's rule 
 * divide [ptime, ctime] into [ptime, ptime + stime / 3.0]
 * [ptime + stime / 3.0, ptime + 2.0 * stime / 3.0]
 * [ptime + 2.0 * stime / 3.0, ctime]
 * use Simpson's rule separately
 */
static FLOAT
estimate_source_error(DOF *f_h)
{
    GRID *g = f_h->g;
    DOF *tmp;
    FLOAT result = 0.0, time, h;

    time = ctime;
    h = stime / 6.0;
   
    /* P:1/3 */
    tmp = phgDofNew(g, DOF_DEFAULT, 1, "tmp", DofNoAction);
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result += phgDofNormL2(tmp);

    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result += 4.0 * phgDofNormL2(tmp);
    
    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result +=  2.0 * phgDofNormL2(tmp);
    
    /* M: 1/ 3 */
    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result +=  4.0 * phgDofNormL2(tmp);
    
    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result += 2.0 *  phgDofNormL2(tmp);
   
    /* L:1/3 */
    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result += 4.0 * phgDofNormL2(tmp);
    
    ctime -= h;
    phgDofSetDataByFunction(tmp, func_f);
    phgDofAXPY(-1.0, f_h, &tmp);
    result += phgDofNormL2(tmp);
    phgDofFree(&tmp);
    
    ctime = time;
    return result / 18.0;
}

static FLOAT  /* space error indicator*/
estimate_space_error(DOF *u_h, DOF *u_p, DOF *f_h, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *tmp, *jump;
    DOF *p_g, *h_g, *div;
    FLOAT eta, h;
    int i;
  
    tmp = phgDofGradient(u_h, NULL, NULL, "tmp");
    jump = phgQuadFaceJump(tmp, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    phgDofFree(&tmp);
    eta = 1.0 / stime;
    tmp = phgDofCopy(f_h, NULL, DOF_DGn[u_h->type->order], NULL);
    phgDofAXPY(eta, u_p, &tmp);
    eta *= -1.0;
    phgDofAXPY(eta, u_h, &tmp);
    p_g = phgDofGradient(u_p, NULL, NULL, NULL);
    h_g = phgDofGradient(u_h, NULL, NULL, NULL);
    div = phgDofDivergence(p_g, NULL, NULL, NULL);
    phgDofAXPY(-0.5, div, &tmp);
    phgDofFree(&div);
    phgDofFree(&p_g);
    div = phgDofDivergence(h_g, NULL, NULL, NULL);
    phgDofAXPY(-0.5, div, &tmp);
    phgDofFree(&div);
    phgDofFree(&h_g);

    ForAllElements(g, e){
        eta = 0.0;
        for (i = 0; i < NFace; i++) {
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;    /* boundary face */
                h = phgGeomGetFaceDiameter(g, e, i);
                eta +=  (*DofFaceData(jump, e->faces[i])) * h;
        }
        h = phgGeomGetDiameter(g, e);
        eta += eta * 0.5  + h * h * phgQuadDofDotDof(e, tmp, tmp, QUAD_DEFAULT);

        *DofElementData(error, e->index) = eta;
    }
    phgDofFree(&tmp);
    phgDofFree(&jump);
    return phgDofNormInftyVec(error);
}

static FLOAT  /* data oscillation */
estimate_osc(DOF *f_h, DOF *u_h, DOF *u_p, DOF *osc)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *res, *cons;
    double pk, tmp;
    double h, v;
    
    tmp = 1 / stime;
    res = phgDofCopy(f_h, NULL, DOF_DEFAULT, "res");
    phgDofAXPY(tmp, u_p, &res);
    tmp *= -1.0;
    phgDofAXPY(tmp, u_h, &res);

    cons = phgDofNew(g, DOF_DEFAULT, 1, "cons", DofNoAction);
    phgDofSetDataByValue(cons, 1.0);
    
    ForAllElements(g, e){
        h = phgGeomGetDiameter(g, e);
	v = phgGeomGetVolume(g, e);
	pk = phgQuadDofDotDof(e, res, cons, -1) / v;
	tmp = (phgQuadDofDotDof(e, res, res, -1) - pk * pk * v) * h * h;
	*DofElementData(osc, e->index) = tmp;
    }
    phgDofFree(&res);
    phgDofFree(&cons);
    return phgDofNormInftyVec(osc);
}

/* calculate inital error */
static FLOAT
Init_error(DOF *u_h, DOF *u_t, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    FLOAT tmp;
    
    ForAllElements(g, e){
        tmp = 0.0;
	tmp += phgQuadDofDotDof(e, u_h, u_h, 4);
	tmp -= 2.0 * phgQuadDofDotDof(e, u_h, u_t, 4);
	tmp += phgQuadDofDotDof(e, u_t, u_t, 4);
	*DofElementData(error, e->index) = tmp;
    }

    return phgDofNormInftyVec(error);
}

/* calculate L2 norm of (u_t - u_h) */
/* u_t's type is DOF_ANALYTIC */
static FLOAT
L2_error(DOF *u, DOF *u_h)
{
    SIMPLEX *e;
    FLOAT tmp, result;

    tmp = 0.0;
    ForAllElements(u->g, e){
	tmp += phgQuadDofDotDof(e, u_h, u_h, 5);
	tmp -= 2.0 * phgQuadDofDotDof(e, u_h, u, 5);
	tmp += phgQuadDofDotDof(e, u, u, 5);
    }
    result = tmp;
    return Sqrt(result);
}

/* cal L2 norm for DOF_ANALYTIC */
static FLOAT
L2_norm(DOF *u)  
{
    SIMPLEX *e;
    FLOAT tmp, result;

    tmp = 0.0;
    ForAllElements(u->g, e){
	tmp += phgQuadDofDotDof(e, u, u, 5);
    }
    result = tmp;
    return Sqrt(result);
}

static void
result_print(DOF *u_t, DOF *u_h, DOF *error, 
          FLOAT PE_error, FLOAT etaspace, FLOAT etaf, FLOAT etatime)
{
    size_t peakmem;
    FLOAT tmp, tmp1;
    
    phgPrintf("time step : %lg\n", (double)stime);
    phgPrintf("source error:  %lg          tol:%lg\n",
				(double)etaf, (double)FTOL);
    phgPrintf("time residual: %lg          tol:%lg\n",
				(double)etatime, (double)TimeTOL);
    
    phgPrintf("max space error:       %lg\n", (double)PE_error);
    phgPrintf("space error indicator: %lg  tol:%lg\n",
				(double)etaspace, (double)SpaceTOL);
	
    tmp = L2_norm(u_t);
    tmp1 = L2_error(u_t, u_h);
    phgPrintf("true solution norm:      %0.8lf\n", (double)tmp);
    phgPrintf("numerical solution norm: %0.8lf\n", (double)L2_norm(u_h));
    phgPrintf("True Error:              %lg\n", (double)tmp1);
    phgPrintf("relative Error:          %lg\n", (double)(tmp1/tmp));
    phgMemoryUsage(u_h->g, &peakmem);
    phgPrintf("memory usage: %lf Mb\n", peakmem / (1024. * 1024.));
}

int
main(int argc, char *argv[])
{

    GRID *g;
    SIMPLEX *e;
    DOF *u_h, *u_p, *u_t, *error, *osc, *ftmp, *f_h;
    double tt[3], tt1[3], tt2[3];
    FLOAT etaf, etatime, etaspace;  /* error indicator */
    FLOAT PE_error;           /* max space indicator */
    FLOAT tmp, tmp1;
    FLOAT delta1 = 0.5, delta2 = 2, thetatime = 0.5;  /* time-step size control parameters */
    char *fn = "../test/cube.dat";
    int flag = 0;
    long long Nelem = 0; /* total elements */
    long long Nmesh = 0; /* total DOF */
    char *sn = NULL;     /* output file name */

    phgInit(&argc, &argv); 
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, TRUE))
        phgError(1, "can't read file \"%s\".\n", fn);

    sn = (char *)phgAlloc(30 * sizeof(char));
    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    u_t = phgDofNew(g, DOF_ANALYTIC, 1, "u_t", func_g);
    u_p = phgDofNew(g, DOF_DEFAULT, 1, "u_p", func_u0);
    f_h = phgDofNew(g, DOF_DEFAULT, 1, "f_h",  DofNoAction);
    ftmp = phgDofNew(g, DOF_DEFAULT, 1, "ftmp", DofNoAction);
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);
    osc = phgDofNew(g, DOF_P0, 1, "oscillaiton", DofNoAction);

    FTOL = Sqrt(tol_time) / (2.0 * T);
    phgRefineAllElements(g, 4);
    phgBalanceGrid(g, 1.02, -1, NULL, 0.);
    phgGetTime(tt);
    phgPrintf("Number of processes:%d\n", phgNProcs);

    /* init U0 such that L2 norm (u0 - U0) < Sqrt(tol_space) */
    etaspace = 2 * tol_space;
    while(etaspace > tol_space){
	phgRefineMarkedElements(g);

	PE_error = Init_error(u_p, u_t, error);
	etaspace = phgDofNormL1Vec(error);
	phgPrintf("Init grid. Error:%lg    tol:%lg\n",
				(double)etaspace, (double)tol_space);
	phgPrintf("Number of DOF:           %d\n", DofGetDataCountGlobal(u_h));
	phgPrintf("Number of elements:      %d\n\n", g->nleaf_global);
		
	tmp = 0.25 * PE_error;
	ForAllElements(g, e){
            e->mark = 0;
	    if (*DofElementData(error, e->index) > tmp)
		e->mark = 1;
	}
    }

    while(ctime < T - 1e-8){
        phgGetTime(tt1);
        ptime = ctime;
        ctime += stime;
	phgPrintf("\n/********* start new time layer *************/\n");
	phgPrintf("current time layer: [%lf, %lf]\n",
				(double)ptime, (double)ctime);

	if (ctime > T){
	    ctime = T;
	    stime = T - ptime;
	    phgPrintf("current time layer: [%lf, %lf]\n",
				(double)ptime, (double)ctime);
	}

	flag++;
	if (flag > 1){   /* update u_p */
	    phgDofFree(&u_p);
            u_p = phgDofCopy(u_h, NULL, DOF_DEFAULT, "u_p2");
	    u_p->userfunc = DofInterpolation;
	}

	/* update f_h */
        update_fh(f_h, ftmp);
        
	SolveLinearSystem(SOLVER_DEFAULT, u_h, f_h, u_p);
	
	/* calculate error indicators */
        etaf = estimate_source_error(f_h);
	etatime = estimate_time_error(u_h, u_p);
        PE_error = estimate_space_error(u_h, u_p, f_h, error);
	etaspace = phgDofNormL1Vec(error);
	
        /* output result */
        result_print(u_t, u_h, error, PE_error, etaspace, etaf, etatime);
	
	/* time step control */
	while(etaf > FTOL || etatime > TimeTOL){
            phgPrintf("\n>>>>>>>>>>>>>>> adjust time step <<<<<<<<<<<<\n");
            stime *= delta1;
            ctime = stime + ptime;

            phgPrintf("time step: %lg\n", (double)stime);
	    phgPrintf("current time layer: [%lf, %lf]\n",
				(double)ptime, (double)ctime);

	    /* update f_h */
            update_fh(f_h, ftmp);
	    
	    SolveLinearSystem(SOLVER_DEFAULT, u_h, f_h, u_p);
            
            /* calculate errors */
	    etaf = estimate_source_error(f_h);
            etatime = estimate_time_error(u_h, u_p);
            PE_error = estimate_space_error(u_h, u_p, f_h, error);
	    etaspace = phgDofNormL1Vec(error);
	
            /* output result */
            result_print(u_t, u_h, error, PE_error, etaspace, etaf, etatime);
	}
	
	phgPrintf("\ntime step: %lg\n\n", (double)stime);

	while(etaspace > SpaceTOL){
            /* refine and coarsen elements */
            phgPrintf("\n=============== adjust grid =================\n");
            estimate_osc(f_h, u_h, u_p, osc);
#if 1
	    MarkElements(error, osc);
#else
	    phgMarkElements_(MARK_MAX, error, refine_max, osc, osc_max, \
			  1, 2, MARK_MAX, error, coarsen_max, 1, 2);
#endif
	    phgPrintf("refine elements: ");
	    cal_time(FALSE);
	    phgRefineMarkedElements(g);
	    cal_time(TRUE);
	    phgPrintf("\ncoarsen elements:");
	    cal_time(FALSE);
	    phgCoarsenMarkedElements(g);
	    cal_time(TRUE);
	    phgPrintf("\n\n");

	    if (phgBalanceGrid(g, 1.02, -1, NULL, 0.))
                phgPrintf("Repartition mesh, lif:   %lg\n\n", (double)g->lif);

	    /* updat f_h */
            update_fh(f_h, ftmp);

	    SolveLinearSystem(SOLVER_DEFAULT, u_h, f_h, u_p);

            /* calculate errors */
            etaf = estimate_source_error(f_h);
            etatime = estimate_time_error(u_h, u_p);
            PE_error = estimate_space_error(u_h, u_p, f_h, error);
	    etaspace = phgDofNormL1Vec(error);
	
            /* output result */
            result_print(u_t, u_h, error, PE_error, etaspace, etaf, etatime);

	    /* time step control */
	    while(etaf > FTOL || etatime > TimeTOL){
                phgPrintf("\n>>>>>>>>>>>>>>>>adjust time step <<<<<<<<<<<<<\n");
                stime *= delta1;
                ctime = stime + ptime;
                phgPrintf("time step: %lg\n", (double)stime);
	        phgPrintf("current time: [%le, %le]\n",
				(double)ptime, (double)ctime);

		/* update f_h */
                update_fh(f_h, ftmp);

	        SolveLinearSystem(SOLVER_DEFAULT, u_h, f_h, u_p);
	   
		/* cal error indicators */
                etaf = estimate_source_error(f_h);
                etatime = estimate_time_error(u_h, u_p);
                PE_error = estimate_space_error(u_h, u_p, f_h, error);
	        etaspace = phgDofNormL1Vec(error);
	
                /* output result */
                result_print(u_t, u_h, error, PE_error, etaspace, etaf, etatime);
	    }
	    phgPrintf("max error:%lg   etaspace:%lg  tol:%lg\n",
			(double)PE_error, (double)etaspace, (double)SpaceTOL);
	}
	
	phgDofSetDataByFunction(ftmp, func_g);
	phgDofAXPY(-1.0, u_h, &ftmp);
#if OUTPUT_MESH
	*sn = '\0';
	if (flag % 40 == 0){
            sprintf(sn, "heat-4-%5.5d.vtk", flag / 40);
	    phgPrintf("output mesh to %s:\n", sn);
            phgExportVTK(g, sn, u_h, NULL);
	}
#endif

	tmp = thetatime * tol_time;
	tmp1 = Sqrt(tmp) / (2.0 * T);
	tmp = thetatime * TimeTOL;

	/* increase time step if time error indicator is too small */
	if (etatime <= tmp && etaf <= tmp1){
            stime *= delta2;
	    phgPrintf("adjust time-step to: %lf\n", (double)stime);
	}

	Nelem += g->nleaf_global;
	Nmesh += DofGetDataCountGlobal(u_h);
	phgGetTime(tt2);
	phgPrintf("\ntime usage of this step: %lfs\n",
				(double)(tt2[2] - tt1[2]));
    }

    /* cal total errors */
    phgPrintf("\nTotal time steps:%d\n", flag);
    phgPrintf("\nTotal elements:%ld\n", Nelem);
    Nelem /= flag;
    phgPrintf("Average elements:%ld\n", Nelem);
    phgPrintf("\nTotal DOF:%ld\n", Nmesh);
    Nmesh /= flag;
    phgPrintf("Average DOF:%ld\n", Nmesh);
    phgGetTime(tt2);
    phgPrintf("\nTotal time:%lfs\n", (double)(tt2[2] - tt[2]));

    free(sn);
    phgDofFree(&ftmp);
    phgDofFree(&osc);
    phgDofFree(&u_h);
    phgDofFree(&u_p);
    phgDofFree(&f_h);
    phgDofFree(&u_t);
    phgDofFree(&error);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
