/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: maxwell-eigen.c,v 1.20 2007/09/27 07:37:20 zlb Exp $
 *
 * Compute eigenvalues of time harmonic Maxwell's equations */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define LAGRANGE_MULTIPLIER	1

static void
build_matrices(MAT *matA, MAT *matM, DOF *u_h)
{
    int N = u_h->type->nbas;	/* number of basis functions in an element */
    int i, j;
    GRID *g = u_h->g;
    SIMPLEX *e;
    FLOAT A[N][N], B[N][N], buffer[N], d;
    INT I[N];
#if LAGRANGE_MULTIPLIER
    int order = u_h->type->order;
    DOF_TYPE *type = DOF_Pn[order];
    DOF *p_h = phgDofNew(u_h->g, type, 1, "p_h", DofNoAction);
    int M = p_h->type->nbas;
    FLOAT C[M][N];
    INT k;
#endif	/* LAGRANGE_MULTIPLIER */

    assert(DofTypeDim(u_h->type) == Dim);

    assert(u_h->dim == 1);
    ForAllElements(g, e) {
	for (i = 0; i < N; i++) {
	    I[i] = phgMapE2L(matA->rmap, 0, e, i);
#if LAGRANGE_MULTIPLIER
	    for (k = 0; k < M; k++) {
		/* \int \grad\psi_k\cdot\phi_i */
		C[k][i] =
		    phgQuadGradBasDotBas(e, p_h, k, u_h, i, QUAD_DEFAULT);
	    }
#endif	/* LAGRANGE_MULTIPLIER */
	    for (j = 0; j <= i; j++) {
		/* \int \phi_i\cdot\phi_j */
		B[j][i] = B[i][j] = 
		    phgQuadBasDotBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
		d = 0.;
#if LAGRANGE_MULTIPLIER
		for (k = 0; k < M; k++)
		    d += C[k][i] * C[k][j];
#endif	/* LAGRANGE_MULTIPLIER */
		/* \int \curl\phi_i\cdot\curl\phi_j */
		A[j][i] = A[i][j] = -d +
		    phgQuadCurlBasDotCurlBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	    }
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (!phgDofDirichletBC(u_h, e, i, NULL, buffer, NULL,
					DOF_PROJ_CROSS)) {
		phgMatAddEntries(matA, 1, I + i, N, I, A[i]); 
		phgMatAddEntries(matM, 1, I + i, N, I, B[i]); 
	    }
	}
    }
#if LAGRANGE_MULTIPLIER
    phgDofFree(&p_h);
#endif	/* LAGRANGE_MULTIPLIER */
}

static FLOAT
estimate_error(FLOAT lambda, DOF *u_h, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *curl_u_h, *jump1, *jump2, *residual, *tmp;

    curl_u_h = phgDofCurl(u_h, NULL, NULL, NULL);
    residual = phgDofCopy(u_h, NULL, DOF_DGn[u_h->type->order], "residual");
    tmp = phgDofCurl(curl_u_h, NULL, NULL, NULL);
    phgDofAXPBY(-1.0, tmp, lambda, &residual);
    phgDofFree(&tmp);
    jump1 = phgQuadFaceJump(curl_u_h, DOF_PROJ_CROSS, NULL, QUAD_DEFAULT);
    jump2 = phgQuadFaceJump(u_h, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    ForAllElements(g, e) {
	int i;
	FLOAT eta, h;
	FLOAT diam = phgGeomGetDiameter(g, e);
	e->mark = 0;		/* clear refinement mmark */
	eta = 0.0;
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] == DIRICHLET || e->bound_type[i] == NEUMANN)
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    eta += (*DofFaceData(jump1, e->faces[i]) +
		    *DofFaceData(jump2, e->faces[i])) * h;
	}
	eta = eta + diam * diam * phgQuadDofDotDof(e, residual, residual,
						   QUAD_DEFAULT);
	*DofElementData(error, e->index) = Sqrt(eta);
    }
    phgDofFree(&jump1);
    phgDofFree(&jump2);
    phgDofFree(&curl_u_h);
    phgDofFree(&residual);
    return phgDofNormInftyVec(error);
}

int
main(int argc, char *argv[])
{
#if 0
    /* \lambda_lnm = \pi^2(l^2+m^2+n^2) with lm+ln+nm > 0 */
    static char *fn = "../test/cube.dat";
    INT nev = 20;
#else
    /* j_m(\sqrt{\lambda}) == 0 (TE) or j'_m(\sqrt{\lambda'}) == 0 (TM),
     * where j_m is the m-th order spherical Bessel function, j'_m its
     * derivative */
    static char *fn = "../test/sphere.dat";
    INT nev = 20;
#endif

    INT pre_refines = 0;
    size_t mem, mem_peak;
    int i, j, n, nit;
    GRID *g;
    SIMPLEX *e;
    DOF **u_h, **error;
    MAP *map;
    MAT *A, *M;
    FLOAT tol = 1e-2;
    FLOAT PEoo, thres;
    FLOAT *evals;
    FLOAT wtime;

    phgOptionsPreset("-default_dof_type ND1");
    phgOptionsRegisterInt("nev", "Number of eigen pairs", &nev);
    phgOptionsRegisterInt("pre_refines", "Pre_refines", &pre_refines);
    phgOptionsRegisterFloat("tol", "Tolerance", &tol);
    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    /* pre-refinement */
    phgRefineAllElements(g, pre_refines);

    evals = phgCalloc(nev, sizeof(*evals));
    u_h = phgAlloc(nev * sizeof(*u_h));
    error = phgAlloc(nev * sizeof(*error));
    for (i = 0; i < nev; i++) {
	u_h[i] = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
	phgDofSetDirichletBoundaryMask(u_h[i], BDRY_MASK);
	/* set random initial values for the eigenvectors */
	phgDofRandomize(u_h[i], i == 0 ? 123 : 0);
	error[i] = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);
    }

    while (TRUE) {
	phgPrintf("\n");
	if (phgBalanceGrid(g, 1.2, -1, NULL, 0.))
	    phgPrintf("Repartition mesh, load imbalance: %lg\n",(double)g->lif);
	phgPrintf("%d DOF, %d elements, %d submesh%s load imbalance: %lg\n",
			DofGetDataCountGlobal(u_h[0]), g->nleaf_global,
			g->nprocs, g->nprocs > 1 ? "es" : "", (double)g->lif);
	map = phgMapCreate(u_h[0], NULL);
	A = phgMapCreateMat(map, map);
	A->assembled = FALSE;
	M = phgMapCreateMat(map, map);
	M->assembled = FALSE;
	build_matrices(A, M, u_h[0]);
	phgMatRemoveBoundaryEntries(A);
	phgMatRemoveBoundaryEntries(M);
	wtime = phgGetTime(NULL);
#if LAGRANGE_MULTIPLIER
	n = phgMatEigenSolve(A, M, nev, EIGEN_CLOSEST, 20.0, &nit, evals,
				map, u_h, NULL);
#else	/* LAGRANGE_MULTIPLIER */
	n = phgMatEigenSolve(M, A, nev, EIGEN_LARGEST, 0.0, &nit, evals,
				map, u_h, NULL);
	phgMatDestroy(&M);
	phgMatDestroy(&A);
	phgMapDestroy(&map);
	for (i = 0, j = n - 1; i < j; i++, j--) {
	    DOF *tmp;
	    tmp = u_h[n - 1 - i];
	    u_h[n - 1 - i] = u_h[i];
	    u_h[i] = tmp;
	    PEoo = evals[n - 1 - i];
	    evals[n - 1 - i] = evals[i];
	    evals[i] = PEoo;
	}
	for (i = 0; i < n; i++)
	    evals[i] = 1.0 / evals[i];
#endif	/* LAGRANGE_MULTIPLIER */
	for (i = n; i < nev; i++)
	    phgDofRandomize(u_h[i], 0);
	phgPrintf("%d iterations, converged eigenvalues: %d, wtime: %0.2lfs\n",
			nit, n, (double)(phgGetTime(NULL) - wtime));
	for (i = 0, j = 0; i < n; i++) {
	    /* remove negative eigen values */
	    if (evals[i] < 1e-5)
		continue;
	    if (j != i) {
		evals[j] = evals[i];
		phgDofCopy(u_h[i], u_h + j, NULL, NULL);
	    }
	    phgPrintf("  tau[%d]=%0.6le, L2(u_h[%d])=%0.6le\n",
		j, (double)evals[j], j, (double)phgDofNormL2(u_h[j]));
	    j++;
	}
	n = j;

	PEoo = -1.0;
	for (i = 0; i < n; i++) {
	    thres = estimate_error(evals[i], u_h[i], error[i]);
	    if (PEoo < thres)
		PEoo = thres;
	}
	PEoo = Fabs(PEoo);
	phgPrintf("indicator=%0.3le\n", (double)PEoo);
	mem = phgMemoryUsage(g, &mem_peak);
	phgPrintf("Memory usage: current %0.4lgMB, peak %0.4lgMB\n",
		(double)mem / (1024.0 * 1024.0),
		(double)mem_peak / (1024.0 * 1024.0));
	if (PEoo <= tol  || mem_peak >= 1024 * (size_t)1024 * 400)
	    break;
	if (n == 0) {
	    phgRefineAllElements(g, 1);
	    continue;
	}
	ForAllElements(g, e)
	    e->mark = 0;
	/*thres = 1e-2 / Sqrt((double)g->nleaf_global);*/
	thres = PEoo * 0.5;
	for (i = 0; i < n; i++)
	    ForAllElements(g, e)
		if (*DofElementData(error[i],e->index) >= thres)
		    e->mark = 1;
	phgRefineMarkedElements(g);
    }

#if 1
    phgPrintf("Creating \"%s\".\n", phgExportVTKn(g, "output.vtk", n, u_h));
#endif

    phgFree(evals);
    for (i = 0; i < nev; i++) {
	phgDofFree(u_h + i);
	phgDofFree(error + i);
    }
    phgFree(u_h);
    phgFree(error);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
