/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* This file implements marking strategies MAX, MNS and GERS
 *
 * Author: Liu Hui (liuhui@lsec.cc.ac.cn)
 *
 * $Id: mark.c,v 1.7 2007/09/27 07:37:20 zlb Exp $ */

#include "phg.h"

#include <math.h>

static int phgMarkRecur;  

/*  info will be output if phgVerbosity >= phgMarkInfo */
static int phgMarkInfo = 0;

/******************** determine factor ********/
static FLOAT
dof_factor_sum(DOF *x, FLOAT mx, FLOAT factor, int flag, int con)
{
    GRID *g = x->g;
    SIMPLEX *e;
    INT i;
    FLOAT sum, b;
    FLOAT tmp, a;

    assert(factor >= 0. && mx >= 0. && factor <= 1.);
    assert((flag == 1 && con == 0) ||
           (flag == 0 && con == 0) ||
           (flag == 0 && con == 1) ||
	   (flag == -1) || (flag == -2));

    tmp = mx * factor;
    /******************************* refine ****************************/
    
    if (flag == 1 && con == 0){  /* face data and all e->marks are 0*/
        sum = 0.0;
        ForAllElements(g, e){
            for (i = 0; i < NFace; i++){
                if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                    continue;
                b = *DofFaceData(x, e->faces[i]);
               if (b > tmp){
                   sum += b;
	       }
	    }
	}

        return sum * 0.5;
    }
    
    if (flag == 0 && con == 0){    /* element data && e->mark == 0*/
        sum = 0.0;
	ForAllElements(g, e){
            b = *DofElementData(x, e->index);
            if (b > tmp){
                sum += b;
	    }
	}
        return sum;
    }

    /* con = 1 and flag = 0, data are stored in element */
    /* some marks are depth */
    if (flag == 0 && con == 1){
        b = 0.0;
        ForAllElements(g, e){
            if (e->mark >= 1){
                b += *DofElementData(x, e->index);
	    }
        }
        sum = b;

        a = 0.0;
        ForAllElements(g, e){
            if (e->mark == 0){
                b = *DofElementData(x, e->index);
		if (b > tmp){
                    a += b;
		}
	    }
        }
        sum += a;
        return sum;
    }

    /***************************** coarsen ****************************/
    if (flag == -1){  /* face data */
        sum = 0.0;
        ForAllElements(g, e){
            for (i = 0; i < NFace; i++){
                if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                    continue;
                b = *DofFaceData(x, e->faces[i]);
               if (b < tmp){
                   sum += b;
	       }
	    }
	}
        return sum * 0.5;
    }
    
    /* element data */
    sum = 0.0;
    ForAllElements(g, e){
        b = *DofElementData(x, e->index);
        if (b < tmp){
            sum += b;
	}
    }
    return sum;
}

static FLOAT
mark_recursion(DOF *error, FLOAT mx, FLOAT eta, 
               FLOAT low, FLOAT up, int flag, int con)
{
    FLOAT sm;
    FLOAT c1, c2, c3;
    static int tmp = 0;
    static FLOAT low_sum, up_sum;

    phgMarkRecur++;
    assert(low >= 0.0 && up >= low && up <= 1.0);
    assert(mx >= 0. && eta >= 0.);

    /*************************  refine ****************/
    if (flag >= 0){
	tmp++;
        sm = (low + up) * 0.5;
	if (tmp > 1){
            c2 = dof_factor_sum(error, mx, sm, flag, con);
	}
	else {
            c1 = dof_factor_sum(error, mx, low, flag, con);
            c2 = dof_factor_sum(error, mx, sm, flag, con);
            c3 = dof_factor_sum(error, mx, up, flag, con);
	    low_sum = c1;
	    up_sum = c3;
	}
	if (low_sum == c2 || c2 == up_sum || c2 == eta){	/* end */
            tmp = 0;
            return sm;
	}

	if (c2 > eta ){
            low_sum = c2;
            return mark_recursion(error, mx, eta, sm, up, flag, con);
	}
        else {
            up_sum = c2;
            return mark_recursion(error, mx, eta, low, sm, flag, con);
	}
    }

    /************************  coarsen *********************/
    sm = (low + up) * 0.5;
 
    if (tmp > 1){
        c2 = dof_factor_sum(error, mx, sm, flag, con);
    }
    else {
        c1 = dof_factor_sum(error, mx, low, flag, con);
        c2 = dof_factor_sum(error, mx, sm, flag, con);
        c3 = dof_factor_sum(error, mx, up, flag, con);
	low_sum = c1;
	up_sum = c2;
    }
    
    if (low_sum == c2 || c2 == up_sum || c2 == eta){	/* end */
	tmp = 0;
        return low;
    }

    if (c2 < eta ){
	low_sum = c2;
        return mark_recursion(error, mx, eta, sm, up, flag, con);
    }
    else {
	up_sum = c2;
        return mark_recursion(error, mx, eta, low, sm, flag, con);
    }
}
/********************  finish determining factor ********/

/***************************************************************/
/*                     refine subroutine                       */
/*                                                             */
/***************************************************************/

static long int /* GERS: mark elements only uses data in faces */
mark_refine_by_facedata(DOF *error, FLOAT fac, int  depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT tmpc;
    FLOAT tmp = 0.;
    FLOAT gama;
    FLOAT sum, flag, PE_error;
    int i;
    long count, cc;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 1 && error->type->np_elem == 0);
    assert(depth > 0);

    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("<<<<< GERS: Marked by FaceData\n");
    
    ForAllElements(g, e){
        e->mark = 0;
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    PE_error = phgDofNormInftyVec(error);
    if (PE_error <= 0.0)
        return 0;
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, PE_error, flag, 0.0, 1.0, 1, 0);
    sum = dof_factor_sum(error, PE_error, gama, 1, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark */
    tmpc = PE_error * gama;
    ForAllElements(g, e){
        for (i = 0; i < NFace; i++){
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;
            tmp = *DofFaceData(error, e->faces[i]);
	    if (tmp > tmpc){
                e->mark = depth;
	    }
	}
    }

    /* number of elements marked */
    cc = 0;
    ForAllElements(g, e){
        if (e->mark == depth)
            cc++;
    }
     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

static long int  /* GERS: mark elements only uses element data */
mark_refine_by_elemdata(DOF *error, FLOAT fac, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT gama;
    FLOAT sum, flag, pe;
    long cc, count;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 0 && error->type->np_elem == 1);
    assert(depth > 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf(">>>>> GERS: Marked by ElementData\n");
    ForAllElements(g, e){
        e->mark = 0;
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    pe = phgDofNormInftyVec(error);
    if (pe <= 0.0)
        return 0;
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, pe, flag, 0.0, 1.0, 0, 0);
    sum = dof_factor_sum(error, pe, gama, 0, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark */
    pe *= gama;
    cc = 0;
    ForAllElements(g, e){
        gama = *DofElementData(error, e->index);
        if (gama > pe){
            e->mark = depth;
            cc++;
	}
    }

    /* number of elements marked */
     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

/* GERS: mark elements uses error in faces */
/* and data oscillation in elements */
static long int
mark_refine_by_MNS(DOF *error, FLOAT fac, DOF *osc, FLOAT fac2, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT gama;
    FLOAT sum, flag, PE_error;
    int i;
    long count, cc;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 1 && error->type->np_elem == 0);
    assert(osc->type->np_vert == 0 && osc->type->np_edge == 0 &&
           osc->type->np_face == 0 && osc->type->np_elem == 1);
    assert(depth > 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("<<<<<< GERS: Marked by MNS strategy\n");
    ForAllElements(g, e){
        e->mark = 0;
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    PE_error = phgDofNormInftyVec(error);
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, PE_error, flag, 0.0, 1.0, 1, 0);
    sum = dof_factor_sum(error, PE_error, gama, 1, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark by error indicator */
    sum = PE_error * gama;
    ForAllElements(g, e){
        for (i = 0; i < NFace; i++){
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;
            gama = *DofFaceData(error, e->faces[i]);
	    if (gama > sum){
                e->mark = depth;
	    }
	}
    }

    /* mark by data oscillation */
    phgMarkRecur = -1;
    PE_error = phgDofNormInftyVec(osc);
    if (PE_error <= 0.0) 
        goto stat;
    flag = phgDofNormL1Vec(osc) * fac2;

    /* determine factor */
    sum = dof_factor_sum(osc, PE_error, 1.0, 0, 1);
    if (sum >= flag)
        goto stat;
    gama = mark_recursion(osc, PE_error, flag, 0.0, 1.0, 0, 1);
    sum = dof_factor_sum(osc, PE_error, gama, 0, 1);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("*****Marked by Data oscillation\n");
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    sum = PE_error * gama;
    ForAllElements(g, e){
        gama = *DofElementData(osc, e->index);
	if (e->mark == 0 && gama > sum){
            e->mark = depth;
	}
    }
    /* number of elements marked */
stat:
    cc = 0;
    ForAllElements(g, e){
        if (e->mark == depth)
            cc++;
    }
     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

/* GERS: mark elements uses error in elements */
/* and data oscillation in elements */
static long int
mark_refine_by_GERS(DOF *error, FLOAT fac, DOF *osc, FLOAT fac2, int depth)
{
    GRID *g = osc->g;
    SIMPLEX *e;
    FLOAT gama, pe;
    FLOAT sum, flag;
    long cc, count;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 0 && error->type->np_elem == 1);
    assert(osc->type->np_vert == 0 && osc->type->np_edge == 0 &&
           osc->type->np_face == 0 && osc->type->np_elem == 1);
    assert(depth > 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf(">>>>> Marked by GERS strategy\n");
    ForAllElements(g, e){
        e->mark = 0;
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    pe = phgDofNormInftyVec(error);
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, pe, flag, 0.0, 1.0, 0, 0);
    sum = dof_factor_sum(error, pe, gama, 0, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark by error indicator */
    pe *= gama;
    ForAllElements(g, e){
        gama = *DofElementData(error, e->index);
        if (gama > pe){
            e->mark = depth;
	}
    }

    /* mark by data oscillation */
    /* determine factor */
    phgMarkRecur = -1;
    pe = phgDofNormInftyVec(osc);
    if (pe == 0.)
        goto stat;
    flag = phgDofNormL1Vec(osc) * fac2;
    sum = dof_factor_sum(osc, pe, 1.0, 0, 1);
    if (sum >= flag)
        goto stat;
    gama = mark_recursion(osc, pe, flag, 0.0, 1.0, 0, 1);
    sum = dof_factor_sum(osc, pe, gama, 0, 1);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("Marked by Data oscillation\n");
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    pe *= gama;
    ForAllElements(g, e){
        gama = *DofElementData(osc, e->index);
        if (e->mark == 0 && gama > pe){
            e->mark = depth;
	}
    }

    /* number of elements marked */
stat:
    cc = 0;
    ForAllElements(g, e){
	if (e->mark == depth)
            cc++;
    }
     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

static long int  /* element data */
mark_refine_by_max(DOF *error, FLOAT theta, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT tmp;
    long int cc, count;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 0 && error->type->np_elem == 1);
    assert(depth > 0);

    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("Marked By Max strategy\n");
    /* max strategy */
    tmp = phgDofNormInftyVec(error) * theta;
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("max:%.8lg  flag:%.8lg  factor:%.8lg\n",
		(double)phgDofNormInftyVec(error), (double)tmp, (double)theta);
    cc = count = 0;
    ForAllElements(g, e){
        e->mark = 0;
        if (*DofElementData(error, e->index) > tmp){
            e->mark = depth;
	    cc++;
	}
    }
    count = cc;
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("Elements marked for refining:%ld\n", count);
    return count;
}

/******************************************************************
 * 
 *                   coarsen subroutine 
 *  if the subroutines are used just followinng the refining 
 *     subroutines, flag should be zero.
 *
 *****************************************************************/

static long int /* GERS: mark elements only uses data in faces */
mark_coarsen_by_facedata(DOF *error, FLOAT fac, unsigned cont, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT tmpc;
    FLOAT tmp = 0.;
    FLOAT gama;
    FLOAT sum, flag, PE_error;
    int i;
    long count, cc;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 1 && error->type->np_elem == 0);
    assert(depth < 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("<<<<< GERS: Marked by FaceData\n");
    if (cont != 0){
        ForAllElements(g, e){
            e->mark = 0;
	}
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    PE_error = phgDofNormInftyVec(error);
    if (PE_error <= 0.)
        return 0;
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, PE_error, flag, 0.0, 1.0, -1, 0);
    sum = dof_factor_sum(error, PE_error, gama, -1, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark */
    tmpc = PE_error * gama;
    ForAllElements(g, e){
        for (i = 0; i < NFace; i++){
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;
            tmp = *DofFaceData(error, e->faces[i]);
	    if (tmp < tmpc){
                e->mark = depth;
	    }
	}
    }

    /* number of elements marked */
    cc = 0;
    ForAllElements(g, e){
        if (e->mark == depth)
            cc++;
    }
     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

static long int  /* GERS: mark elements only uses element data */
mark_coarsen_by_elemdata(DOF *error, FLOAT fac, unsigned cont, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT gama;
    FLOAT sum, flag, pe;
    long cc, count;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 0 && error->type->np_elem == 1);
    assert(depth < 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf(">>>>> GERS: Marked by ElementData\n");
    if (cont != 0){
        ForAllElements(g, e){
            e->mark = 0;
	}
    }
   
    /* determine factor */
    phgMarkRecur = -1;
    pe = phgDofNormInftyVec(error);
    if (pe <= 0.)
        return 0;
    flag = phgDofNormL1Vec(error) * fac;

    gama = mark_recursion(error, pe, flag, 0.0, 1.0, -2, 0);
    sum = dof_factor_sum(error, pe, gama, -2, 0);
    if (phgVerbosity >= phgMarkInfo){
        phgPrintf("recursion times:%d\n", phgMarkRecur);
        phgPrintf("flag:%.8lg  mark:%.8lg  factor:%lf\n",
			(double)flag, (double)sum, (double)gama);
    }
    
    /* mark */
    pe *= gama;
    cc = 0;
    ForAllElements(g, e){
        gama = *DofElementData(error, e->index);
        if (gama < pe){
            e->mark = depth;
            cc++;
	}
    }

     count = cc;
     if (phgVerbosity >= phgMarkInfo)
         phgPrintf("Elements marked for refining:%ld\n", count);
     return count;
}

static long int /* max strategy */
mark_coarsen_by_max(DOF *error, FLOAT theta, unsigned int con, int depth)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT tmp;
    long int cc, count;
    
    assert(error->type->np_vert == 0 && error->type->np_edge == 0 &&
           error->type->np_face == 0 && error->type->np_elem == 1);
    assert(depth < 0);
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("Marked By Max strategy\n");
    if (con != 0)
        ForAllElements(g, e){
            e->mark = 0;
	}

    /* max strategy */
    tmp = phgDofNormInftyVec(error) * theta;
    cc = count = 0;
    ForAllElements(g, e){
        if (*DofElementData(error, e->index) < tmp && e->mark == 0){
            e->mark = depth;
	    cc++;
	}
    }
    count = cc;
    if (phgVerbosity >= phgMarkInfo)
        phgPrintf("Elements marked for coarsening: %ld\n", count);
    return count;
}

long int
phgMarkElements_(int strategy, DOF *error, FLOAT theta, DOF *osc,
			FLOAT zeta, int depth, FLOAT p,
		 int coarsen_strategy, DOF *coarsen_error, FLOAT gamma,
			int coarsen_depth, FLOAT coarsen_p)
{
    static BOOLEAN initialized = FALSE;
    static const int strategy_list[] = {MARK_MAX,MARK_GERS,MARK_MNS,MARK_ALL};
    static const char *strategy_names[] = {"max", "gers", "mns", "all", NULL};
    static int default_strategy = 1;
    /* FIXME: do we need to add default_coarsen_strategy? */

    GRID *g;
    SIMPLEX *e;
    FLOAT re, da, co;
    long int cc = -1;
    
    if (!initialized){
	initialized = TRUE;
	phgOptionsRegisterTitle("\nMarking strategy options:", NULL, "marking");
	phgOptionsRegisterKeyword("default_strategy", "Default strategy",
				strategy_names, &default_strategy);
	return 0;
    }

    assert(p >= 1.);
    assert(coarsen_p >= 1.);
    assert(depth > 0);
    assert(coarsen_depth > 0);
    assert(error != NULL);
    assert(strategy == MARK_ALL ||
	   error->type->np_vert + error->type->np_edge +
           error->type->np_face + error->type->np_elem == 1);
    assert(osc == NULL ||
	   osc->type->np_vert + osc->type->np_edge +
           osc->type->np_face + osc->type->np_elem == 1);
    assert(coarsen_error == NULL ||
	   coarsen_error->type->np_vert + coarsen_error->type->np_edge +
           coarsen_error->type->np_face + coarsen_error->type->np_elem == 1);

    if (strategy == MARK_DEFAULT)
	strategy = strategy_list[default_strategy];

    re = Pow(theta, p);
    da = Pow(zeta, p);

    /* mark elements for refining */

    switch (strategy) {
	case MARK_ALL:
	    g = error->g;
	    ForAllElements(g, e)
		e->mark = depth;
	    return g->nleaf_global;
	case MARK_MAX:
	    cc = mark_refine_by_max(error, re, depth);
	    break;
	case MARK_MNS:
	    cc = mark_refine_by_MNS(error, re, osc, da, depth);
	    break;
	case MARK_GERS:
	    if (error->type->np_elem == 1) {
		if (osc == NULL)
		    cc =  mark_refine_by_elemdata(error, re, depth);
		else
		    cc = mark_refine_by_GERS(error, re, osc, da, depth);
	    }
	    else {
		cc = mark_refine_by_facedata(error, re, depth);
	    }
	    break;
	default:
	    phgError(1, "phgMarkElements: invalid refine strategy %d!\n",
								strategy);
    }
    
    /* mark elements for coarsening */

    if (coarsen_strategy == MARK_NONE)
	return cc;
    else if (coarsen_strategy == MARK_DEFAULT)
	coarsen_strategy = strategy_list[default_strategy];

    co = Pow(gamma, coarsen_p);

    switch (coarsen_strategy) {
	case MARK_MAX:
	    cc += mark_coarsen_by_max(coarsen_error, co, 0, -coarsen_depth);
	    break;
	case MARK_GERS:
	    if (coarsen_error->type->np_face == 1)
		cc += mark_coarsen_by_facedata(coarsen_error, co, 0,
							-coarsen_depth);
	    else
		cc += mark_coarsen_by_elemdata(coarsen_error, co, 0,
							-coarsen_depth);
	    break;
	default:
	    phgError(1,"phgMarkElements: invalid coarsen strategy %d!\n",
							coarsen_strategy);
    }

    return cc;
}
