/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-mumps.c,v 1.55 2007/09/27 07:37:20 zlb Exp $ */

/* The MUMPS interface (http://mumps.enseeiht.fr/) */

#define _GNU_SOURCE	/* for feenableexcept() */

#include "phg.h"

#if USE_MUMPS	/* whole file */

#define Initialize		NULL
#define Finalize		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

#include <mpi.h>

#include <dmumps_c.h>
#include <stdlib.h>
#include <string.h>

#if HAVE_FEENABLEEXCEPT
# include <fenv.h>
#endif  /* HAVE_FEENABLEEXCEPT */

/* macros s.t. indices match documentation */
#define ICNTL(I)	icntl[(I)-1]
#define CNTL(I)		cntl[(I)-1]
#define INFO(I)		info[(I)-1]
#define INFOG(I)	infog[(I)-1]

typedef struct {
    DMUMPS_STRUC_C	*mumps_id;
    BOOLEAN		factored;

    int			symmetry;
    INT			refine;
    INT			fillin;
    char		*filename;
#if HAVE_FEENABLEEXCEPT
    BOOLEAN		disable_fpetrap;
#endif  /* HAVE_FEENABLEEXCEPT */
#if 0
    int			maxis;
    int			maxs;
#endif
} OEM_DATA;

#define	_t	((OEM_DATA *)solver->oem_data)
#define id	(_t->mumps_id)

static const char *symmetry_list[] = {"unsym", "spd", "sym", NULL};
static int symmetry = 2;	/* symmetry of the matrix, defaults to sym */
static INT refine = 0;		/* max nbr of iterative refinements */

static INT fillin = 20;		/* percentage of extra fillin. Note: MUMPS's
				 * built-in default is 15 (SPD) and 20 (gen.) */

static char *filename = NULL;	/* filename to write the matrix */

#if HAVE_FEENABLEEXCEPT
static BOOLEAN disable_fpetrap = TRUE;
#endif  /* HAVE_FEENABLEEXCEPT */
#if 0
static int maxis = 0;		/* workspace size for integers */
static int maxs = 0;		/* workspace size for doubles */
#endif

static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nMUMPS options:", NULL, "mumps");
    phgOptionsRegisterKeyword("mumps_symmetry", "Symmetry of the matrix",
				symmetry_list, &symmetry);
    phgOptionsRegisterInt("mumps_refine", "Max number of iterative "
				"refinements", &refine);
    phgOptionsRegisterInt("mumps_fillin", "Extra fill-in (percent, "
				"< 0 means use default setting)", &fillin);
#if HAVE_FEENABLEEXCEPT
    phgOptionsRegisterNoArg("mumps_disable_fpetrap", "Disable FPE trap",
				&disable_fpetrap);
#endif  /* HAVE_FEENABLEEXCEPT */

    phgOptionsRegisterFilename("mumps_dump_matrix", "Dump the matrix",
				&filename);

#if 0
    phgOptionsRegisterInt("mumps_maxis", "Workspace size for integers", &maxis);
    phgOptionsRegisterInt("mumps_maxs", "Workspace size for doubles", &maxs);
#endif

    return 0;
}

static void
call_dmumps(SOLVER *solver)
{
#if HAVE_FEENABLEEXCEPT
    int saved_excepts = 0;
    if (_t->disable_fpetrap)
	saved_excepts = fedisableexcept(FE_UNDERFLOW | FE_DIVBYZERO |
					FE_INVALID | FE_OVERFLOW);
#endif  /* HAVE_FEENABLEEXCEPT */

    /* call mumps */
    dmumps_c(id);

#if HAVE_FEENABLEEXCEPT
    if (_t->disable_fpetrap) {
	feclearexcept(saved_excepts);
	feenableexcept(saved_excepts);
    }
#endif  /* HAVE_FEENABLEEXCEPT */
}

static int
Init(SOLVER *solver)
{
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));

    _t->symmetry	= symmetry;
    _t->refine		= refine;
    _t->fillin		= fillin;
    if (filename != NULL)
	_t->filename	= strdup(filename);
#if HAVE_FEENABLEEXCEPT
    _t->disable_fpetrap	= disable_fpetrap;
#endif  /* HAVE_FEENABLEEXCEPT */
#if 0
    _t->maxis		= maxis;
    _t->maxs		= maxs;
#endif

    return 0;
}

static int
Create(SOLVER *solver)
{
    assert(solver->mat->type != PHG_MATRIX_FREE);

    id = phgCalloc(1, sizeof(DMUMPS_STRUC_C));
    _t->factored = FALSE;

    phgInfo(2, "create MUMPS solver.\n");

    id->par = 1;
    id->sym = _t->symmetry;
#ifdef DONT_HAVE_G_COMM
    id->comm_fortran = MPI_Comm_c2f(MPI_COMM_WORLD);
#else	/* DONT_HAVE_G_COMM */
    id->comm_fortran = MPI_Comm_c2f(solver->mat->rmap->comm);
#endif	/* DONT_HAVE_G_COMM */
    id->job = -1;
    call_dmumps(solver);

    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL || id == NULL)
	return 0;

    phgInfo(2, "destroy MUMPS solver.\n");

    phgFree(id->irn_loc);
    phgFree(id->jcn_loc);
    phgFree(id->a_loc);
    if (solver->mat->rmap->rank == 0)
	phgFree(id->rhs);

    id->job = -2;
    call_dmumps(solver);
    phgFree(id);
    phgFree(_t->filename);
    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    MAP *map = solver->mat->rmap;
    int i, j, k, nz;
    int Nlocal = map->nlocal, n0 = map->partition[map->rank];
    MAT_ROW *row;
    double time0;

    if (_t->factored)
	return 0;

    if (solver->sent_to_oem == TRUE)
	phgError(1, "%s:%d, data already sent to OEM solver!\n",
						__FILE__, __LINE__);
    solver->sent_to_oem = TRUE;

    assert(solver->mat->type != PHG_DESTROYED);

    if (solver->mat->type == PHG_PACKED)
	phgMatUnpack(solver->mat);

    phgInfo(2, "send data to MUMPS solver.\n");

    /* count # of nonzeros */
    nz = 0;
    if (_t->symmetry == 0) {
	for (i = 0, row = solver->mat->rows; i < Nlocal; i++)
	    nz += (row++)->ncols;
    }
    else {
	for (i = 0, row = solver->mat->rows; i < Nlocal; i++, row++)
	    for (j = 0; j < row->ncols; j++) {
		if ((k = row->cols[j]) < i)
		    continue;
#ifndef DONT_HAVE_G_COMM
		if (k >= Nlocal && solver->mat->O2Gmap[k - Nlocal] < n0)
		    continue;
#endif	/* DONT_HAVE_G_COMM */
		    nz++;
	    }
    }
    id->nz_loc = nz;

    id->irn_loc = phgAlloc(nz * sizeof(*id->irn_loc));
    id->jcn_loc = phgAlloc(nz * sizeof(*id->jcn_loc));
    id->a_loc = phgAlloc(nz * sizeof(*id->a_loc));

    if (map->rank == 0) {
	id->n = solver->mat->rmap->nglobal;
	id->nrhs = 1;
	id->lrhs = solver->mat->rmap->nglobal;
	id->rhs = phgAlloc(id->n * sizeof(*id->rhs));
    }

    /* setup MUMPS matrix */
    nz = 0;
    for (i = 0, row = solver->mat->rows; i < Nlocal; i++, row++) {
	for (j = 0; j < row->ncols; j++) {
	    k = row->cols[j];
#ifndef DONT_HAVE_G_COMM
	    k = (k < Nlocal ? k + n0 : solver->mat->O2Gmap[k - Nlocal]);
#endif	/* DONT_HAVE_G_COMM */
	    if (_t->symmetry != 0 && k < n0 + i)
		continue;
	    id->a_loc[nz] = row->data[j];
	    id->irn_loc[nz] = 1 + n0 + i;
	    id->jcn_loc[nz++] = 1 + k;
	}
	/* free matrix row in solver */
	if (solver->mat->ref_count == 0) {
	    phgFree(row->cols);
	    phgFree(row->data);
	    row->cols = NULL;
	    row->data = NULL;
	    row->ncols = row->alloc = 0;
	}
    }

    if (solver->mat->ref_count == 0)
	phgMatFreeMatrix(solver->mat);

    if (_t->filename != NULL) {
	/* dump matrix to fn.rank.mat */
	static int no = 0;
	char *fn = phgAlloc(strlen(_t->filename) + 32);
	FILE *fp;
	if (no == 0 && map->rank == 0) {
	    sprintf(fn, "/bin/rm -f %s[0-9][0-9].[0-9]*.mat", _t->filename);
	    system(fn);
	}
	sprintf(fn , "%s%02d.%02d.mat", _t->filename, no++, map->rank);
	phgInfo(1, "saving matrix to \"%s\".\n", fn);
	if ((fp = fopen(fn, "w+t")) == NULL)
	    phgError(1, "cannot open output file \"%s\".\n", fn);
	phgFree(fn);
	fprintf(fp, "# global matrix size, symmetry, nproc\n");
	fprintf(fp, "%d %d %d\n", map->nglobal, _t->symmetry, map->nprocs);
	fprintf(fp, "# number of local nonzero entries\n");
	fprintf(fp, "%d\n", id->nz_loc);
	fprintf(fp, "# entries: i j a_ij\n");
	for (i = 0; i < id->nz_loc; i++)
	    fprintf(fp, "%d %d %0.17lg\n", id->irn_loc[i], id->jcn_loc[i],
			(double)id->a_loc[i]);
	fclose(fp);
    }

    _t->factored = TRUE;

    phgInfo(2, "analyse and factor matrix.\n");

    /* set MUMPS parameters */
    /* No outputs */
    if (phgVerbosity <= 0) {
	id->ICNTL(1) = -1;
	id->ICNTL(2) = -1;
	id->ICNTL(3) = -1;
    }
    id->ICNTL(4) = phgVerbosity;
    id->ICNTL(18) = 3;			/* matrix in distributed format */

    /* call MUMPS analyser (JOB 1 = analyse) */
    time0 = phgGetTime(NULL);
    id->job = 1;
    call_dmumps(solver);
    phgInfo(1, "Time for Analysis: %lg\n", (double)(phgGetTime(NULL) - time0));
#if 0
    if (_t->maxis > 0)
	id->maxis = _t->maxis;
    if (_t->maxs > 0)
	id->maxs = _t->maxs;
#endif
    if (_t->fillin >= 0)
	id->ICNTL(14) = _t->fillin;	/* extra fill-in percentage */
    id->CNTL(2) = solver->rtol;		/* stopping criterion for refinement */
    id->ICNTL(10) = _t->refine;		/* number of iterative refinements */
    time0 = phgGetTime(NULL);
    id->job = 2;			/* factorize */
    call_dmumps(solver);
    phgInfo(1, "Time for Factorize: %lg\n", (double)(phgGetTime(NULL) - time0));

    if ((i = id->INFOG(1)) < 0) {
	if (map->rank == 0) {
	    switch (i) {
		case -8:  case -9:  case -11: case -12:
		case -14: case -15: case -18: case -19:
		    phgInfo(-1, "\n\n");
		    phgInfo(-1, "The return code of MUMPS is %d, which means "
			       "an error has occurred in\n", i);
		    phgInfo(-1, "the solution phase (the value of "
			       "'-mumps_fillin' is %d, you may try\n",
				id->ICNTL(14));
		    phgInfo(-1, "rerun the program with a larger value).\n");
		    break;
		case -5: case -7: case -13:
		    phgInfo(-1, "\n\n");
		    phgInfo(-1, "The return code of MUMPS is %d, which means "
			       "a failed alloc (retry with\n", i);
		    phgInfo(-1, "a smaller value of -mumps_fillin?)\n");
		    break;
		default:
		    phgInfo(-1, "\n\n");
		    phgInfo(-1, "The return code of MUMPS is %d\n", i);
		    break;
	    }
	}
	phgError(1, "abort.\n");
    }
    else if (i > 0 && map->rank == 0) {
	phgWarning("MUMPS returns a non-zero code: %d\n", i);
    }

    return 0;
}
 
static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    MAP *map = solver->mat->rmap;
    int i, Nlocal = map->nlocal;
    double time0;

    Assemble(solver);

    /* gather RHS data */
    for (i = 0; i < Nlocal; i++)
	id->rhs[i] = solver->rhs->data[i];

    time0 = phgGetTime(NULL);
    id->job = 3;		/* solve */
    call_dmumps(solver);
    phgInfo(1, "Time for Solve: %lg\n", (double)(phgGetTime(NULL) - time0));

    if (destroy) {
	phgFree(id->irn_loc);
	phgFree(id->jcn_loc);
	id->irn_loc = id->jcn_loc = NULL;
	phgFree(id->a_loc);
	id->a_loc = NULL;
    }

    /* Scatter solution */
    for (i = 0; i < Nlocal; i++)
	x->data[i] = id->rhs[i];

    if (destroy && map->rank == 0) {
	phgFree(id->rhs);
	id->rhs = NULL;
    }

    solver->residual = 0.0;	/* fake residual norm */
    return solver->nits = 1;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverMUMPS_ = {
    "MUMPS", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy,
    AddMatrixEntries, AddRHSEntries,
    Assemble, Solve
};

#endif	/* USE_MUMPS */
